/* engine.cpp
 *
 * ZEthereal -- Ethereal for the Sharp Zaurus
 * Copyright (C) 2003  Philippe Biondi <biondi@cartel-securite.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */



#include <iostream.h>
#include "engine.h"

extern "C" {
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include "../config.h"
#include <malloc.h>
#include <string.h>
#include <pcap.h>
#include "../pcap-util.h"
#include "engine_core.h"
#include "../epan/resolv.h"
}


Engine::Engine(void)
{
	childpid = 0;
	count = 0;
	promisc = 0;
	resolv_flags = 0;
	filter = NULL;
	rfilter = NULL;
	interface = NULL;
	filename = NULL;		
	engine_core_init();
	childpid=0;
	forked=0;
}

Engine::~Engine(void)
{
	kill_child();
	if (filter) free(filter);
	if (rfilter) free(rfilter);
	if (interface) free(interface);
	if (filename) free(filename);
}

int Engine::is_main_instance(void)
{
	return (forked == 0);
}


void Engine::kill_child(void)
{
	if (childpid)
		kill(childpid, SIGTERM);
	childpid = 0;
}

GList *Engine::get_interfaces(void)
{
	int err;
	char err_str[64];

	return  get_interface_list(&err, err_str);
}

void Engine::free_interfaces(GList *l)
{
	free_interface_list(l);
}


void Engine::set_count(int c) { capture_opts.autostop_count = c; }
void Engine::set_filter(const char *f) { if (filter) { free(filter); } filter = strdup(f); }
void Engine::set_rfilter(const char *rf) { if (rfilter) { free(rfilter); } rfilter = strdup(rf); }
void Engine::set_interface(const char *i) { if (cfile.iface) { g_free(cfile.iface); } cfile.iface = g_strdup(i); }
void Engine::set_filename(const char *f) { if (filename) { free(filename); } filename = strdup(f); }
void Engine::set_promisc(int p) { capture_opts.promisc_mode = p; }
void Engine::set_resolv_flags(int dns, int mac, int port) 
{ 
	g_resolv_flags = ( (dns ? RESOLV_NETWORK : 0) | 
			 (mac ? RESOLV_MAC : 0 ) |
			 (port ? RESOLV_TRANSPORT : 0) );
}

void Engine::do_capture()
{
	dump_state("Capture");

	int pid;
	pid = fork();
	switch (pid) {
	case 0:
		forked = 1;
		break;
	case -1:
		cerr << "do_capture: fork: error " << errno << endl;
		break;
	default:
		childpid = pid;
		cout << "{{{{" << cfile.iface << endl;
		engine_core_start(NULL, filter);
		kill_child();
		waitpid(childpid, NULL, 0);
	}
}

void Engine::do_stop_capture(void)
{
	if (forked) {
		forked=0;
		kill(getppid(), SIGUSR1);
		exit(0);
	}
}

void Engine::do_read_file()
{
	dump_state("Read file");
	engine_core_start(filename, rfilter);
}

void Engine::dump_state(char *title = "")
{
	if (title)
		cout << title << endl;

	cout << "Count=[" << capture_opts.autostop_count << "]" << endl;
	cout << "filter=[" << filter << "]" << endl;
	cout << "rfilter=[" << rfilter << "]" << endl;
	cout << "interface=[" << cfile.iface << "]" << endl;
	cout << "filename=[" << filename << "]" << endl;
	cout << "promisc=[" << capture_opts.promisc_mode << "]" << endl;
	cout << "resolv_flags=[" << g_resolv_flags << "]" << endl;
}
